import 'package:audio_video_progress_bar/audio_video_progress_bar.dart';
import 'package:fis_measure/interfaces/process/player/play_controller.dart';
import 'package:fis_measure/view/player/controller.dart';
import 'package:fis_measure/view/player/events.dart';
import 'package:fis_vid/data_channel/progress_info.dart';
import 'package:flutter/material.dart';
import 'package:get/get.dart';
import 'package:fis_measure/view/mobile_view/widgets/throttle.dart' as utils;

class StreamingProgressBarWithCrossFrame extends StatefulWidget {
  const StreamingProgressBarWithCrossFrame({super.key});

  @override
  State<StatefulWidget> createState() => _ProgressBarState();
}

class _ProgressBarState extends State<StreamingProgressBarWithCrossFrame> {
  late VidPlayerController controller;

  int totalCount = 0;
  int currentIndex = 0;
  double bufferedProgress = 0;

  int get bufferedCount {
    int count = (bufferedProgress * totalCount).toInt();
    final currentPlayedIndex = controller.currentFrame?.index ?? 0;
    if (count < currentPlayedIndex) {
      // 缓冲进度至少需要追平播放(可播放)进度
      count = currentPlayedIndex;
    }
    return count;
  }

  late List<int> anchors;

  @override
  void initState() {
    _initController();
    super.initState();
  }

  @override
  void dispose() {
    _removeControllerListeners();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return Stack(
      children: [
        Center(
          child: buildBufferBar(),
        ),
        buildCrossFrameAnchorBar(),
        buildLocationBar(),
      ],
    );
  }

  Widget buildBufferBar() {
    // 仅用于显示加载进度
    return ProgressBar(
      progress: Duration.zero,
      total: Duration(milliseconds: totalCount),
      buffered: Duration(milliseconds: bufferedCount),
      // bufferedBarColor: Colors.red,
      barHeight: 12,
      thumbRadius: 10,
      thumbColor: Colors.transparent,
      // baseBarColor: Colors.purple,
      timeLabelLocation: TimeLabelLocation.none,
    );
  }

  Widget buildLocationBar() {
    return SliderTheme(
      data: SliderThemeData(
        trackShape: _NoPaddingCustomTrackShape(),
        trackHeight: 12,
        thumbShape: const RoundSliderThumbShape(enabledThumbRadius: 10),
        thumbColor: Colors.white,
        overlayShape: const RoundSliderOverlayShape(overlayRadius: 0),
        overlayColor: Colors.transparent,
        // activeTrackColor: Colors.black,
        // inactiveTrackColor: Colors.transparent,
      ),
      child: Slider(
        value: currentIndex.toDouble(),
        min: 0,
        max: totalCount.toDouble(),
        onChanged: (v) {
          final index = v.round();
          if (index == currentIndex) {
            return;
          }
          setState(() {
            currentIndex = index;
          });
          utils.throttle(() {
            controller.locateTo(currentIndex);
          }, 'onDragTo', 80);
        },
        onChangeStart: (v) {
          //手动拖动时移除事件监听【避免重复setstate】
          controller.eventHandler.removeListener(_onControllerEvent);
        },
        onChangeEnd: (v) {
          controller.eventHandler.addListener(_onControllerEvent);
        },
      ),
    );
  }

  Widget buildCrossFrameAnchorBar() {
    return LayoutBuilder(builder: (context, c) {
      return RepaintBoundary(
        key: ValueKey(controller.hashCode),
        child: CustomPaint(
          painter: _AnchorBarPainter(totalCount, anchors),
          size: Size(c.maxWidth, c.maxHeight),
        ),
      );
    });
  }

  void _update() {
    /// 第一帧下标为0 ，所以 totalFramesCount 要减1
    totalCount = controller.totalFramesCount - 1;
    currentIndex = controller.currentFrameIndex;
    if (mounted) {
      setState(() {});
    }
  }

  void _onDownloadProgressChanged(Object sender, VidDownloadProgressInfo e) {
    if (e.progress == 1) {
      bufferedProgress = 1;
    } else {
      bufferedProgress = _calcRealFramesBufferPercent();
    }
    if (mounted) {
      setState(() {});
    }
  }

  void _onControllerEvent(Object sender, VidPlayerEvent e) {
    if (e is VidPlayerFrameIndexChangeEvent) {
      if (currentIndex != e.index) {
        _update();
      }
    }
  }

  void _onCrossFrameAnchorsUpdated(Object sender, List<int> e) {
    setState(() {
      anchors = e;
    });
  }

  void _onControllerDisposeOccured(Object sender, void e) {
    _removeControllerListeners();
    if (mounted) {
      setState(() {
        anchors = [];
      });
    }
  }

  void _initController() {
    controller = Get.find<IPlayerController>() as VidPlayerController;
    anchors = controller.crossFrameAnchors;
    WidgetsBinding.instance.addPostFrameCallback((timeStamp) {
      _addControllerListeners();
    });
  }

  void _addControllerListeners() {
    controller.eventHandler.addListener(_onControllerEvent);
    controller.dataChannel.downloadProgressChanged
        .addListener(_onDownloadProgressChanged);
    controller.crossFrameAnchorsUpdated
        .addListener(_onCrossFrameAnchorsUpdated);
    // controller.disposeOccured.addListener(_onControllerDisposeOccured);
  }

  void _removeControllerListeners() {
    controller.dataChannel.downloadProgressChanged
        .removeListener(_onDownloadProgressChanged);
    controller.eventHandler.removeListener(_onControllerEvent);
    controller.crossFrameAnchorsUpdated
        .removeListener(_onCrossFrameAnchorsUpdated);
    // controller.disposeOccured.removeListener(_onControllerDisposeOccured);
  }

  double _calcRealFramesBufferPercent() {
    final channel = controller.dataChannel;
    final int fileSize = channel.getFileSize();
    final int bufferredSize = channel.getBufferSize();

    final int framesTotalSize = fileSize - controller.vidHeaderSize;
    final int framesBufferredSize = bufferredSize - controller.vidHeaderSize;
    double percent = framesBufferredSize / framesTotalSize;

    return percent;
  }
}

class _AnchorBarPainter extends CustomPainter {
  static const anchorHeight = 24.0;
  static const anchorWidth = 3.0;
  static const yLocationOffset = anchorHeight / 2;
  static const xLocationOffset = anchorWidth / 2;

  final int totalCount;
  final List<int> anchors;

  _AnchorBarPainter(this.totalCount, this.anchors);

  final painter = Paint()
    ..style = PaintingStyle.fill
    ..color = Colors.yellow;

  double unitWidth = 0.0;
  double centerY = 0.0;
  double topY = 0.0;
  double bottomY = 0.0;

  @override
  void paint(Canvas canvas, Size size) {
    unitWidth = size.width / totalCount;
    centerY = size.height / 2;

    topY = centerY - yLocationOffset;
    bottomY = centerY + yLocationOffset;

    for (var anchor in anchors) {
      drawOne(canvas, anchor);
    }
  }

  void drawOne(Canvas canvas, int anchor) {
    final x = anchor * unitWidth;

    final rrect = RRect.fromLTRBR(
      x - xLocationOffset,
      topY,
      x + xLocationOffset,
      bottomY,
      const Radius.circular(4),
    );

    canvas.drawRRect(rrect, painter);
  }

  @override
  bool shouldRepaint(covariant _AnchorBarPainter oldDelegate) {
    if (oldDelegate.anchors != anchors) {
      return true;
    }
    if (oldDelegate.totalCount != totalCount) {
      return true;
    }
    return false;
  }
}

class _NoPaddingCustomTrackShape extends RoundedRectSliderTrackShape {
  ///去掉默认边距
  @override
  Rect getPreferredRect({
    required RenderBox parentBox,
    Offset offset = Offset.zero,
    required SliderThemeData sliderTheme,
    bool isEnabled = false,
    bool isDiscrete = false,
  }) {
    final double trackHeight = sliderTheme.trackHeight ?? 1;
    // final double trackLeft = offset.dx + trackHeight * .25;
    final double trackLeft = offset.dx;
    final double trackTop =
        offset.dy + (parentBox.size.height - trackHeight) / 2;
    // final double trackWidth = parentBox.size.width - trackHeight * .5;
    final double trackWidth = parentBox.size.width;
    return Rect.fromLTWH(trackLeft, trackTop, trackWidth, trackHeight);
  }
}
